// Program for lecturing purposes by jan ejka (324987@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//

#pragma comment(lib, "devil.lib")

#include <cmath>

// Include GLEW to access OpenGL 4.0 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// Include DevIL for image loading
#include <IL/il.h>

// Use Very Simple Libs
#include "vsMathLib.h"
#include "vsShaderLib.h"

#include "TeapotData.h"

//----------------------------------------------------------------
// TODO:
// 1) Inspect functions "setupShaders", "init", "processKeys" and "renderScene"
// 2) Use SHADER_LEVELS_TRIANGLES, play around with shaders
// 3) Use SHADER_LEVELS_QUADS, play around with shaders
// 4) Use SHADER_LEVELS_ISOLINES, play around with shaders
// 5) Inspect teapot patch in TeapotData.h
// 6 - 12) In teapot shaders
// 13) Play with BUMP_TEX_FILENAME, animate_height_scale, and position of the light
//----------------------------------------------------------------

#define SHADER_LEVELS_TRIANGLES		0
#define SHADER_LEVELS_QUADS			1
#define SHADER_LEVELS_ISOLINES		2
#define SHADER_TEAPOT				3

#define CURRENT_SHADER		SHADER_TEAPOT

//----------------------------------------------------------------

#define BUMP_TEX_FILENAME_1		"textures/Bump1.png"
#define BUMP_TEX_FILENAME_2		"textures/Bump2.png"

#define BUMP_TEX_FILENAME		BUMP_TEX_FILENAME_1

//----------------------------------------------------------------

const bool animate_height_scale = false;

//----------------------------------------------------------------

// Position of the light, in world or camera coordinates according to 'light_in_world'
float lightPos[4] = {0.0f, 0.0f, 0.0f, 1.0f};
float lightAmb[3] = {0.1f, 0.1f, 0.1f};
float lightDif[3] = {1.0f, 1.0f, 1.0f};
float lightSpe[3] = {1.0f, 1.0f, 1.0f};
bool light_in_world = false;

//----------------------------------------------------------------

float materialAmb[3] = {1.0f, 0.0f, 0.0f};
float materialDif[3] = {1.0f, 0.0f, 0.0f};
float materialSpe[3] = {1.0f, 1.0f, 1.0f};
float materialShi = 20.0f;

//----------------------------------------------------------------

VSMathLib *vsml;
VSShaderLib shader;

// Current window size
int win_width, win_height;

// Camera Position
float camX, camY, camZ;

// Mouse Tracking Variables
int startX, startY, tracking = 0;

// Camera Spherical Coordinates
float alpha = 0.0f, beta = 0.0f;
float r = 5.0f;

// Camera offset
float offsetX = 0.0f, offsetY = 0.0f;

// Time counter
float time = 0.0f;

// Textures
GLuint bump_tex;

// Geometry
GLuint vao;
GLuint buffers[2];

bool wireframe = false;
bool increase_time = true;

// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

GLuint loadTexture(std::string filename, GLenum internal_format, bool mipmap)
{
	// Load Texture Map
	unsigned int imageID = 0;
	ilGenImages(1, &imageID);
	ilBindImage(imageID);
	ilEnable(IL_ORIGIN_SET);
	ilOriginFunc(IL_ORIGIN_LOWER_LEFT); 

	ILboolean success = ilLoadImage((const wchar_t *)filename.c_str());
	if (!success)
	{
		// The operation was not sucessfull, hence free image and texture 
		ilDeleteImages(1, &imageID); 
		printf("Couldn't load texture: %s\n", filename.c_str());
		return 0;
	}

	// Add information to the log
	printf("Texture %s succesfully loaded\n", filename.c_str());

	// Convert image to RGBA
	ilConvertImage(IL_RGBA, IL_UNSIGNED_BYTE); 

	// Create and load textures to OpenGL */
	GLuint textureID = 0;
	glGenTextures(1, &textureID);
	glBindTexture(GL_TEXTURE_2D, textureID); 
	glTexImage2D(GL_TEXTURE_2D, 0, internal_format, ilGetInteger(IL_IMAGE_WIDTH), ilGetInteger(IL_IMAGE_HEIGHT), 0,
		GL_RGBA, GL_UNSIGNED_BYTE, ilGetData()); 

	// Mipmapping
	if (mipmap)
		glGenerateMipmap(GL_TEXTURE_2D);

	glBindTexture(GL_TEXTURE_2D, 0);

	// Release the memory
	ilDeleteImages(1, &imageID); 

	return textureID;
}

// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h)
{
	// Prevent a divide by zero, when window is too short
	if(h == 0)
		h = 1;

	win_width = w;
	win_height = h;

	glViewport(0, 0, win_width, win_height);

	// Update the projection matrix of the camera
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, float(win_width) / float(win_height), 0.1f, 10000.0f);

	shader.setUniform("win_width", float(win_width));
	shader.setUniform("win_height", float(win_height));
}


// ------------------------------------------------------------
//
// Render stuff
//

void renderScene()
{
	// Clear the buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Use shader for rendering with shadow
	glUseProgram(shader.getProgramIndex());

	// Compute and set the view matrix
	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->lookAt(camX, camY, camZ, 0,0,0, 0,1,0);
	vsml->translate(VSMathLib::VIEW, offsetX, offsetY, 0.f);

	// Transform light to camera space and send it to GLSL
	if (light_in_world)
	{
		float res[4];
		vsml->multMatrixPoint(VSMathLib::VIEW, lightPos, res);
		shader.setBlockUniform("Lights", "light_position", res);
	}

	shader.setUniform("time", time);
	shader.setUniform("height_scale", 0.3f * (animate_height_scale ? sinf(time / 30.0f) : 1.0f));

	glPolygonMode(GL_FRONT_AND_BACK, wireframe ? GL_LINE : GL_FILL);

	vsml->loadIdentity(VSMathLib::MODEL);

	// Render the scene
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, bump_tex);
	vsml->matricesToGL();
	glBindVertexArray(vao);
#if (CURRENT_SHADER == SHADER_LEVELS_TRIANGLES) || (CURRENT_SHADER == SHADER_LEVELS_QUADS) || (CURRENT_SHADER == SHADER_LEVELS_ISOLINES)
	glEnable(GL_PROGRAM_POINT_SIZE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);		// Always in wireframe
	glPatchParameteri(GL_PATCH_VERTICES, 1);
	glDrawElements(GL_PATCHES, 1, GL_UNSIGNED_INT, nullptr);
#else
	glPatchParameteri(GL_PATCH_VERTICES, 16);
	glDrawElements(GL_PATCHES, TEAPOT_INDICES, GL_UNSIGNED_INT, nullptr);
#endif
	glBindVertexArray(0);
	glBindTexture(GL_TEXTURE_2D, 0);

	// Move the time forward
	if (increase_time)
		time += 1.0f;

	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;
		case 'c':
			printf("Camera Spherical Coordinates (%f, %f, %f)\n", alpha, beta, r);
			break;
		case 'm':
			glEnable(GL_MULTISAMPLE);
			break;
		case 'n':
			glDisable(GL_MULTISAMPLE);
			break;
		case 'w':
			wireframe = !wireframe;
			break;
		case 't':
			increase_time = !increase_time;
			break;
	}

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// ------------------------------------------------------------
//
// Mouse Events
//

void processMouseButtons(int button, int state, int xx, int yy)
{
	// Start tracking the mouse
	if (state == GLUT_DOWN)
	{
		startX = xx;
		startY = yy;
		if (button == GLUT_LEFT_BUTTON)
			tracking = 1;
		else if (button == GLUT_RIGHT_BUTTON)
			tracking = 2;
		else if (button == GLUT_MIDDLE_BUTTON)
			tracking = 3;
	}

	// Stop tracking the mouse
	else if (state == GLUT_UP)
	{
		if (tracking == 1)
		{
			alpha -= (xx - startX);
			beta += (yy - startY);
			if (beta > 85.0f)
				beta = 85.0f;
			else if (beta < -85.0f)
				beta = -85.0f;
		}
		else if (tracking == 2)
		{
			r += (yy - startY) * 0.01f;
			if (r < 0.1f)
				r = 0.1f;
		}
		tracking = 0;
	}
}

// Track mouse motion while buttons are pressed
void processMouseMotion(int xx, int yy)
{
	int deltaX, deltaY;
	float alphaAux, betaAux;
	float rAux;

	deltaX =  - xx + startX;
	deltaY =    yy - startY;

	// Left mouse button: move camera
	if (tracking == 1)
	{
		alphaAux = alpha + deltaX;
		betaAux = beta + deltaY;

		if (betaAux > 85.0f)
			betaAux = 85.0f;
		else if (betaAux < -85.0f)
			betaAux = -85.0f;
		rAux = r;
	}
	// Right mouse button: zoom
	else if (tracking == 2)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r + (deltaY * 0.01f);
		if (rAux < 0.1f)
			rAux = 0.1f;
	}
	// Middle mouse button: move object
	else if (tracking == 3)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r;

		offsetX -= deltaX * 0.0001f;
		offsetY -= deltaY * 0.0001f;
	}

	camX = rAux * sin(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camZ = rAux * cos(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camY = rAux *   						       sin(betaAux * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}

void mouseWheel(int wheel, int direction, int x, int y)
{
	r += direction * 0.1f;
	if (r < 0.1f)
		r = 0.1f;

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// --------------------------------------------------------
//
// Shader Stuff
//

bool setupShaders()
{
	// Shader for shadow usage
	shader.init();
#if CURRENT_SHADER == SHADER_LEVELS_TRIANGLES
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/levels.vert");
	shader.loadShader(VSShaderLib::TESS_CONTROL_SHADER, "shaders/levels_triangles.tesc");
	shader.loadShader(VSShaderLib::TESS_EVAL_SHADER, "shaders/levels_triangles.tese");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/levels.frag");
#elif CURRENT_SHADER == SHADER_LEVELS_QUADS
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/levels.vert");
	shader.loadShader(VSShaderLib::TESS_CONTROL_SHADER, "shaders/levels_quads.tesc");
	shader.loadShader(VSShaderLib::TESS_EVAL_SHADER, "shaders/levels_quads.tese");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/levels.frag");
#elif CURRENT_SHADER == SHADER_LEVELS_ISOLINES
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/levels.vert");
	shader.loadShader(VSShaderLib::TESS_CONTROL_SHADER, "shaders/levels_isolines.tesc");
	shader.loadShader(VSShaderLib::TESS_EVAL_SHADER, "shaders/levels_isolines.tese");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/levels.frag");
#elif CURRENT_SHADER == SHADER_TEAPOT
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/teapot.vert");
	shader.loadShader(VSShaderLib::TESS_CONTROL_SHADER, "shaders/teapot.tesc");
	shader.loadShader(VSShaderLib::TESS_EVAL_SHADER, "shaders/teapot.tese");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/teapot.frag");
#else
#error Invalid CURRENT_SHADER value
#endif

	// Set semantics for the shader variables
	shader.setProgramOutput(0, "outputF");
	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");

	shader.prepareProgram();
	printf("InfoLog for scene shader\n%s\n\n", shader.getAllInfoLogs().c_str());

	shader.setUniform("bumpTex", 0);

	shader.setBlockUniform("Lights", "light_position", lightPos);
	shader.setBlockUniform("Lights", "light_amb", lightAmb);
	shader.setBlockUniform("Lights", "light_dif", lightDif);
	shader.setBlockUniform("Lights", "light_spe", lightSpe);

	shader.setBlockUniform("Material", "mat_amb", materialAmb);
	shader.setBlockUniform("Material", "mat_dif", materialDif);
	shader.setBlockUniform("Material", "mat_spe", materialSpe);
	shader.setBlockUniform("Material", "mat_shi", &materialShi);

	return shader.isProgramLinked();
}


// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

bool init()
{
	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	glGenBuffers(2, buffers);

	glBindBuffer(GL_ARRAY_BUFFER, buffers[0]);
	glBufferData(GL_ARRAY_BUFFER, TEAPOT_VERTICES * 3 * sizeof(float), newellTeapotVertices, GL_STATIC_DRAW);
	glEnableVertexAttribArray(VSShaderLib::VERTEX_COORD_ATTRIB);
	glVertexAttribPointer(VSShaderLib::VERTEX_COORD_ATTRIB, 3, GL_FLOAT, 0, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffers[1]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, TEAPOT_INDICES * sizeof(unsigned int), newellTeapotPatch, GL_STATIC_DRAW);

	glBindVertexArray(0);

	// Texture setup
	ilInit();
	bump_tex = loadTexture(BUMP_TEX_FILENAME, GL_RGB, true);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT); 

	// Some GL settings
	glEnable(GL_DEPTH_TEST);

	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	glClearDepth(1.0f);

	return true;
}

void initVSL()
{
	// Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::MODEL, "m_model");
	vsml->setUniformName(VSMathLib::VIEW, "m_view");
	vsml->setUniformName(VSMathLib::PROJECTION, "m_projection");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA|GLUT_MULTISAMPLE);

	glutInitContextVersion(4, 0);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Tessellation Shaders");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);
	glutMouseFunc(processMouseButtons);
	glutMotionFunc(processMouseMotion);

	glutMouseWheelFunc(mouseWheel);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	int param;
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
	if(param == GL_CONTEXT_CORE_PROFILE_BIT)
		printf("Context Profile: Core\n");
	else
		printf("Context Profile: Compatibility\n");

	if(!setupShaders())
		exit(1);

	initVSL();

	if(!init())
		exit(1);

//  GLUT main loop
	glutMainLoop();

	return(0);
}
